#include "delcher.h"
#include "graph.h"
#include "gene.h"

const int OVLP_SIZE=100000;

struct List
{
  exon *predexon;
  List *link;
};


void freeList(List *L);
void  Process_Options  (int, char * []);
long int printgenes(Site **Sites, long int ssno,int start, int stop, int ignore,int offset,int *offgene);
long int printexon(exon *curexon,int *geneno,int *exonno, long int offset, long int startprint,long int stopprint, int *oktoprint,long int lastseen);
void markexon(struct exon *e,int val);
void freeexon(struct exon *e,int val);
long int datalen(FILE *fp, char Name[]);
long int LoadData(FILE *fp,char *Data, long int chunk, long int goback);

char TRAIN_DIR[500]="";

int main  (int argc, char * argv [])
{
  FILE  * fp;
  char *Data1;
  long int Input_Size;
  long int Data1_Len;
  char  Name1[MAX_LINE];
  long int ssno;
  Site **Sites;
  long int len, proclen,offset,printlen,goback;
  int start,stop;

  // ---------------------------- OPTIONS ------------------------------

  if  (argc < 2) {
    fprintf (stderr,
	     "USAGE:  %s <genome1-file> <training-dir-for-genome1> [options] \n",
	     argv [0]);
    exit (-1);
  }

  Process_Options (argc, argv);   // Set global variables to reflect status of
                                  // command-line options.

  //  -> options should include the possibility to predict in only one genome (maximize score only for one, and not for the other <- kept constant)

  // --------------------------- READ DATA ------------------------------

  fp = File_Open (argv [1], "r");
  
  len=datalen(fp,Name1);

  printf("GlimmerHMM\n");    
  printf("Sequence name: %s\n",Name1); 

  printf("Sequence length: %ld bp\n",len);

  if(len<INIT_SIZE) Input_Size=len+2;
  else Input_Size=INIT_SIZE+2;
  

  Data1 = (char *) malloc(Input_Size*sizeof(char));
  if(Data1 == NULL ) {
      fprintf (stderr, "ERROR:  Unable to alloc memory for input sequence\n");
      exit (0);
  }

  Data1[0]='n';

  printf("\nPredicted genes/exons\n\n");
  printf("Gene Exon Strand  Exon            Exon Range      Exon\n");
  printf("   #    #         Type                           Length\n\n");

  start=1;
  goback=0;
  proclen=0;
  int offgene=1;
  while(proclen<len) {

    Data1_Len = LoadData(fp,Data1,INIT_SIZE,goback)-1;

    // run graph file
    Sites=graph(Data1,Data1_Len,TRAIN_DIR,&ssno);

    // print data
    stop = proclen+Data1_Len<len ? 0 : 1;

    if(!start) offset=proclen-OVLP_SIZE/2;
    else offset=0;

    printlen=printgenes(Sites,ssno,start,stop,OVLP_SIZE/2,offset,&offgene);
    if(!stop) {
      proclen=printlen;
      goback=INIT_SIZE-proclen+offset+OVLP_SIZE/2;
    }
    else proclen=len;
    start=0;
    fprintf(stderr,"Done %d bp\n",proclen);
  }

}


long int LoadData(FILE *fp,char *Data, long int chunk, long int goback) 
{
  int ch;
  long int i;
  
  // find the start
  while(goback) {
    fseek(fp,-1,SEEK_CUR);
    ch=fgetc(fp);
    fseek(fp,-1,SEEK_CUR);
    if(isspace(ch)) continue;
    goback--;
  }

  // copy the data
  Data[0]='n';
  i=1;

  while(i<=chunk && (ch=fgetc(fp))!= EOF && ch !='>') {
    if(isspace(ch)) continue;
    ch = tolower (ch);
    switch  (ch) {
    case  'a' :
    case  'c' :
    case  'g' :
    case  't' :
    case  's' :
    case  'w' :
    case  'r' :
    case  'y' :
    case  'm' :
    case  'k' :
    case  'b' :
    case  'd' :
    case  'h' :
    case  'v' :
    case  'n' :
      break;
    default :
      fprintf (stderr, "Unexpected character `%c\' in input sequence\n",ch);
      ch = 'n';
    }
    Data[i++]=ch;
  }
  Data[i]='\0';

  return(i);
}

long int printgenes(Site **Sites, long int ssno,int start, int stop, int ignore,int offset,int *offgene)
{
  long int i, j, k;
  int ret;
  double maxscore[PREDNO];
  int val, maxval[PREDNO],maxi[PREDNO];
  exon *curexon;
  List *PredExon, *CurExon[PREDNO];
  int geneno,exonno,oktoprint;
  long int startprint,stopprint,lastseen=0;

  for(i=0;i<PREDNO;i++) {
    maxscore[i]=Sites[INTERG][ssno-1].score[i];
    maxval[i]=INTERG;
    maxi[i]=i;
  }

  for(val=I0PLUS; val<=I2MINUS; val++) {
    i=0;
    ret=1;
    j=0;
    while(i<PREDNO && ret) {
      while(j<PREDNO && Sites[val][ssno-1].score[i]<=maxscore[j]) { j++;}
      if(j==PREDNO) ret=0;
      else {
	for(k=PREDNO-1;k>=j+1;k--) {
	  maxscore[k]=maxscore[k-1];
	  maxval[k]=maxval[k-1];
	  maxi[k]=maxi[k-1];
	}
	maxscore[j]=Sites[val][ssno-1].score[i];
	maxval[j]=val;
	maxi[j]=i;
	i++;
	j++;
      }
    }
  }


  for(i=0;i<PREDNO;i++) {
    
    //    printf("Prediction %d:\n",i);

    curexon=Sites[maxval[i]][ssno-1].prevex[maxi[i]];

    if(!stop) 
      while(curexon!=NULL && curexon->type!=2 && curexon->type!=3 && curexon->type!=4 && curexon->type!=7) 
	curexon=curexon->prev;

    PredExon=NULL;

    while(curexon!=NULL) {
      CurExon[i] = (List *) malloc(sizeof(List));
      if (CurExon[i] == NULL) {
	fprintf(stderr,"Memory allocation for current exon failure.\n"); 
	abort();
      }
      CurExon[i]->predexon=curexon;
      CurExon[i]->link=PredExon;
      PredExon=CurExon[i];
      curexon=curexon->prev;
    }

    geneno = *offgene;
    exonno = 1;

    CurExon[i]=PredExon; // I need this later for freeList
    

    if(start) { startprint=0; oktoprint=1;}
    else {startprint=ignore; oktoprint=0;}
    if(stop) { stopprint=INIT_SIZE;}
    else { stopprint=INIT_SIZE-OVLP_SIZE/2;}

    while(PredExon!=NULL) {
      lastseen=printexon(PredExon->predexon,&geneno,&exonno,offset,startprint,stopprint,&oktoprint,lastseen);
      PredExon=PredExon->link;
    }

    *offgene=geneno;
    
  }

  if(!stop) { 
    if(lastseen<INIT_SIZE-OVLP_SIZE/2) lastseen=INIT_SIZE-OVLP_SIZE/2+offset;
    else lastseen+=offset;
  }
  else { lastseen=INIT_SIZE+offset;}

  // ---------------------------- FREE DATA ------------------------------
  
  // free sites 
  for(val=INTERG; val<= I2MINUS; val++) {
    if(Sites[val] != NULL ) {
      for(i=0;i<PREDNO;i++) 
	markexon(Sites[val][ssno-1].prevex[i],val*PREDNO+i);
    }
  }
  for(val=INTERG; val<= I2MINUS; val++) {
    if(Sites[val] !=NULL) {
      for(i=0;i<PREDNO;i++) 
	freeexon(Sites[val][ssno-1].prevex[i],val*PREDNO+i);
    }
    free(Sites[val]);
  }

  // free List
  for(i=0;i<PREDNO;i++) {
    freeList(CurExon[i]);
  }
  
  return(lastseen);

}

void  Process_Options  (int argc, char * argv [])
  
//  Process command-line options and set corresponding global switches
//  and parameters.
//
{
   char  * P;
   long int  W;
   double D;
   int  i;
   FILE  * fp;

   for  (i = 2;  i < argc;  i ++)
     {
      switch  (argv [i] [0])
        {
         case  '-' :
           switch  (argv [i] [1])
             {
	     case 'd' :   //  the training files' directory
	       strcpy(TRAIN_DIR,argv[++i]);strcat(TRAIN_DIR,"/");
	       //printf("train dir = %s\n",TRAIN_DIR);fflush(stdout);
	       break;
	       /*	     case  'm' : // give other PAM/BLOSUM matrix if desired (a 21x21 matrix)
	       fp = fopen(argv[++i],"r");
	       if(fp == NULL) {
	       fprintf(stderr,"ERROR:  Unable to open PAMfile\n");
	       exit(0);
	       }
	       pam=loadpam(fp);
	       usepam=1;
	       fclose(fp);
	       break;
	       case 'p' : // use PAM120 instead of BLOSUM 62
	       pam=pam120;
	       break;
	       case 'b': // use BLOSUM 62 instead of PAM120
	       pam = blosum62;
	       break;*/
	     default :
	       fprintf (stderr, "Unrecognized option %s\n", argv [i]);
             }
           break;
	   /*	case  '+' :
		switch  (argv [i] [1])
		{
		case  'x' :
		// do something else
		break;
		default :
                fprintf (stderr, "Unrecognized option %s\n", argv [i]);
		}
		break;*/
	default :
	  if(i>2) fprintf (stderr, "Unrecognized option %s\n", argv [i]);
	  else strcpy(TRAIN_DIR,argv[i++]);
        }
     }

   return;
}

void freeList(List *L)
{
  if(L==NULL) return;
  freeList(L->link);
  free(L);
  L=NULL;
}


long int printexon(exon *curexon,int *geneno,int *exonno, long int offset, long int startprint,long int stopprint,int *oktoprint,long int lastseen) 
{

  switch(curexon->type) {
  case 0: 
    if(curexon->start<startprint || curexon->start>=stopprint ) { *oktoprint=0;}
    else { *oktoprint=1;}
    if(*oktoprint) {
      printf("%4d %4d  +  Initial  %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
    }
    (*exonno)++;
    break;
  case 1: 
    if(curexon->start<startprint) { *oktoprint=0;}
    if(*oktoprint) printf("%4d %4d  +  Internal %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 2: 
    if(curexon->start<startprint) { *oktoprint=0;}
    if(*oktoprint) {
      printf("%4d %4d  +  Terminal %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
      lastseen=curexon->stop;
      (*geneno)++;
    }
    *exonno=1;
    break;
  case 3: 
    if(curexon->start<startprint || curexon->start>=stopprint ) { *oktoprint=0;}
    else { *oktoprint=1;}
    if(*oktoprint) {
      printf("%4d %4d  +  Single   %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
      lastseen=curexon->stop;
      (*geneno)++;
    }
    *exonno=1;
    break;
  case 4: 
    if(curexon->start<startprint) { *oktoprint=0;}
    if(*oktoprint) {
      printf("%4d %4d  -  Initial  %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
      lastseen=curexon->stop;
      (*geneno)++;
    }
    *exonno=1;
    break;
  case 5: 
    if(curexon->start<startprint) { *oktoprint=0;}
    if(*oktoprint) printf("%4d %4d  -  Internal %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 6: 
    if(curexon->start<startprint || curexon->start>=stopprint ) { *oktoprint=0;}
    else { *oktoprint=1;}
    if(*oktoprint) printf("%4d %4d  -  Terminal %10ld %10ld  %7d\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
    (*exonno)++;
    break;
  case 7: 
    if(curexon->start<startprint || curexon->start>=stopprint ) { *oktoprint=0;}
    else { *oktoprint=1;}
    if(*oktoprint) {
      printf("%4d %4d  -  Single   %10ld %10ld  %7d\n\n",*geneno,*exonno,curexon->start+offset,curexon->stop+offset,curexon->stop-curexon->start+1);
      lastseen=curexon->stop;
      (*geneno)++;
    }
    *exonno=1;
    break;
  }

  return(lastseen);
}
   

void markexon(struct exon *e,int val)
{
  if(e==NULL) return;
  e->type=val;
  markexon(e->prev,val);
}

void freeexon(struct exon *e,int val)
{
  if(e == NULL) return;
  if(e->type>val) return;
  freeexon(e->prev,val);
  if(e->type==val) {
    free(e);
    e=NULL;
  }
}


long int datalen(FILE *fp, char Name[])
{
  char line[MAX_LINE];
  char *P;
  long int len;
  int ch;
  fpos_t pos;

  while((ch=fgetc(fp)) != EOF && ch != '>') ;
  if(ch==EOF) return(0);

  fgets(line, MAX_LINE,fp);
  len=strlen(line);
  assert(len>0 && line[len-1]=='\n');

  fgetpos(fp,&pos);

  P= strtok (line, " \t\n");
  if(P != NULL) strcpy (Name, P);
  else Name [0] = '\0';

  len=0;
  while((ch=fgetc(fp)) != EOF && ch !='>') {
    if(isspace(ch)) continue;
    len++;
  }

  fsetpos(fp,&pos);

  return(len);

}
    
